/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in ShoeKb.java
 * and the file License.html for details.
 */

/* Change History:
   - 10/06/00, JDH: Replaced calls to KBInterface.getQueryVars() with
                    calls to Sentence.getVariables()
   - 05/24/00, JDH: Made member variables private
   - 05/22/00, JDH: Use Query.getNumAnswersRequested to determine how
                    many answers to return in queryKB(), submitQuery()
   - 05/22/00, JDH: Consolidated queryKB() methods
   - 05/19/00, JDH: Added submitSimpleQuery() method
   - 05/19/00, JDH: Made deleteClaims() more efficient
   - 05/19/00, JDH: Removed redundant call to pn.getVariables() from 
                    submitQuery()
   - 05/18/00, JDH: Testing the insertion of an assertion and its claimant
                    assertion in a combined call to server
   */
package ShoeKb;

import ParkaClient.*;
import java.util.*;

/** Implements the KBInterface for dynamic communication with a
  Parka KB. */
public class ParkaInterface extends KBInterface implements ParkaWin {

  private ParkaNet pn;

  private String parkaHost = "";
  private int parkaPort    = -1;

  /* used in a heuristic for determining how large of a string buffer
     if needed for a particular set of claims. */
  private static final int AVG_CLAIM_SIZE = 50;

  public ParkaInterface() {

    pn = new ParkaNet(this);
  }

  /** Establish a connection to the given host and port number. */
  public void establishConnection(String newHost, int newPort)
       throws KBException
  {

    /* if the host or port is changed, update the defaults and
       connect to the new server */
    if (parkaHost.equals(newHost) == false || parkaPort != newPort) {
      parkaHost = newHost;
      parkaPort = newPort;
      pn.changeParkaHost(parkaHost, parkaPort);
    }
    if (pn.isConnection() == false)
      throw new KBException("Couldn't connect to server at " + 
				    parkaHost + ":" + parkaPort);
  }

  /** Creates a new KB with the name supplied by the user. */
  public void createKB(String kbName)
      throws KBException
    {

    }

  /** Initialize the specified knowledge base via connection at the
    given host and port number. */
  public void openKB(String parkaKB, String newHost, int newPort)
       throws KBException 
  {

    establishConnection(newHost, newPort);
    if (pn.isConnection()) {
      pn.openKB(parkaKB, true);
      if (pn.isKBOpen() == false)
	throw new KBException("Could not open KB " + parkaKB);
    }
    /*
    else
      throw new KBException("Couldn't connect to server at " + 
				    parkaHost + ":" + parkaPort);
    */
  }

  /** Open a KB using the existing connection. */
  public void openKB(String parkaKB)
       throws KBException 
  {
    if (pn.isConnection()) {
      pn.openKB(parkaKB, true);
      if (pn.isKBOpen() == false)
	throw new KBException("Could not open KB " + parkaKB);
    } 
    else
      throw new KBException("Cannot open KB. There is no connection.");
  }

  /** Closes the current KB without closing the connection. */
  public void closeKB() {

    pn.closeKB();
  }

  /** Closes the current connection. Any open KBs will be closed as well. */
  public void closeConnection() {

    pn.disconnect();
  }


  /** Issues an arbitrary query to the KB. Variable arguments are 
    preceded by question marks. The answers are returned in a matrix. */
  public ResultSet queryKB(Query query) {

    String[]    varList;
    int         numVars;
    Enumeration enum;
    Atom        atom;
    String[]    args;
    boolean     found;
    String      parkaQuery;
    ResultSet   answers = null;

    /* find which arguments are variables */
    varList = query.getVariables();

    numVars = varList.length;
    if (numVars == 0) {
      System.out.println("ERROR: Query has no variables");
      answers = new ResultSet(null, 0);
    }
    else {
      parkaQuery = formatSentence(query);
      // System.out.println("Submitting query: " + parkaQuery);
      
/* TEST
      System.out.print("vorder: ");
      for (int i=0; i < vorder.length; i++)
	System.out.print(vorder[i] + ", ");
      System.out.println();
*/

      answers = submitQuery(parkaQuery, varList,
			    query.getNumAnswersRequested());
    }
    return answers;
  }


  /** Submits a single conjunct query against the given predicate using the
    given array of arguments. Answers are returned in a two dimensional
    array of variable bindings, where there variables appear
    in the same order as they were in the query. */
  /*
  public ResultSet queryKB(String predicate, String args[]) {

    Query     qry;
    ResultSet answers= null;

    qry = new Query(predicate, args);
    answers = queryKB(qry);
    return answers;
  }
  */

  /** A version of queryKB(String, String[]) for querying binary
    predicates. */
  /*
  public ResultSet queryKB(String predicate, String arg1, String arg2) {

    String[]  args;
    ResultSet answers;

    args = new String[2];
    args[0] = arg1;
    args[1] = arg2;
    answers = queryKB(predicate, args);
    return answers;
  }
  */


  /** Given a Sentence, creates a Parka claims for it. If the Sentence
    has conjuncts, then a claim is created for each conjunct. Two
    inserts are actually made to Parka: one for the actual fact, and one
    that states the claimant of the fact. If the arguments of these
    inserts contain any special Parka characters, the appropriate escape
    codes are inserted. */
  public void storeClaim(String source, Sentence sentence) {

    Vector       atoms;
    StringBuffer assert;
    Enumeration  enum;
    Atom         atom;
    String       parkaAtom;
    String       claim;
    String       claimAssert;

    atoms = sentence.getAtoms();
    // Reserve a reasonable capacity for the StringBuffer
    //   Otherwise its default is 16
    assert = new StringBuffer(atoms.size() * AVG_CLAIM_SIZE);
    enum = atoms.elements();
    while (enum.hasMoreElements()) {
      atom = (Atom)enum.nextElement();
      parkaAtom = formatAtom(atom);
      assert.append(parkaAtom);
      claim = parkaAtom.substring(1, parkaAtom.length()-1);
      claimAssert = formatAssert("claims", source, claim);
      assert.append(claimAssert);
    }
    pn.insertToKB(assert.toString());  
  }

  /** Creates a Parka claim given a source, predicate and array of 
     arguments. The arguments have escape codes inserted if they contain
     any special Parka characters. Two inserts are made to Parka:
     one for the actual fact, and one that states the claimant of 
     the fact. */
  public void storeClaim(String source, String predicate, String args[]) {

    /* NOTE: when we change the system to translate predicates into
       names unique across all ontologies, we must make sure the
       "claims" predicate used below does NOT get translated. */

    String assert;
    String claim;
    String claimAssert;
    int i;

    if (predicate == null)          // do nothing if there's no predicate
      return;
    assert = formatAssert(predicate, args);
    /* keep track of the source of the assertion using claims relation */
    claim = assert.substring(1, assert.length()-1);
    /* formatAssert() properly escapes special characters in all arguments:
       the source may include named anchors */
    claimAssert = formatAssert("claims", source, claim);
    // System.out.println("Storing assertion: " + assert);
    assert = assert + claimAssert;
    // System.out.println("Storing assertion: " + assert);
    pn.insertToKB(assert);  
    // pn.insertToKB(claimAssert);   // inserted with the assert above
  }

  /** A binary version of storeClaim(String, String, String[]) */
  /*
  public void storeClaim(String source, String predicate,
			 String arg1, String arg2) {

    String args[];

    args = new String[2];
    args[0] = arg1;
    args[1] = arg2;
    storeClaim(source, predicate, args);
  }
  */

  /** Deletes all claims made by a source by first getting a list of the
     sources claims, deleting all of these and then checking to see
     if there are any other claims for the same assertions. If not, the
     assertion is deleted as well. */
  /* NOTE: Currently, Parka stores duplicate assertions but a call
     to deleteFromKB() will delete ALL matching assertions. This
     is why we must check if there are other claimants before we delete
     the assertion. */
  public void deleteClaims(String source) {

    /* BUG: since the source is the instance key, if there are 
       multiple pages that have an instance with the same key, all such
       claims from the pages will be deleted.  */
    /* NOTE: when we change the system to translate predicates into
       names unique across all ontologies, we must make sure the
       "claims" predicate used below does NOT get translated. */
    String[]     answers;
    int          numAnswers;
    String       claim;
    String       claimString;
    StringBuffer delBuf;
    int          numClaimants;
    String       delString;
    int          i;

    /* Get a set of the claims made by the source. We use 
       submitSimpleQuery() to avoid the overhead of creating a
       ResultSet. */
    answers = submitSimpleQuery("claims", source, "?X");
    numAnswers = answers.length;
    
    if (numAnswers > 0) {
      /* delete claim information for this source in a single server call */
      // Is Parka doing this correctly?
      delBuf = new StringBuffer(numAnswers * AVG_CLAIM_SIZE);
      System.out.println("Deleting " + numAnswers + " claims");
      for (i=0; i < numAnswers; i++) {
	claim = answers[i];
	claimString = formatAssert("claims", source, claim);
	// pn.deleteFromKB(claimString);
	delBuf.append(claimString);
      }
      // System.out.println("Deleting:\n" + delBuf.toString());
      pn.deleteFromKB(delBuf.toString());

      /* delete the actual assertion if no other sources have claimed it */
      /* we could bundle these all up and make a single delete call, but
	 then we run more of a risk of someone else making the claim
	 between the time we make our query and do the actual deletion.
	 If this happened, we'd delete something that was actually still
	 claimed by somebody. */
      for (i=0; i < numAnswers; i++) {
	claim = answers[i];

	/* find if any other sources have asserted the claim */
	numClaimants = countQueryAnswers(new Query("claims", "?X", claim));
	if (numClaimants <= 0) {
	  delString = "(" + claim + ")";
	  // System.out.println("Deleting " + delString);
	  pn.deleteFromKB(delString);
	  // System.out.println("Status = " + pn.statusMsg);
	}
      }
    }
  }

  /** Creates a new class with the specified name. */
  public void createClass(String className) {

  }

  /** Creates a new predicate in the KB using the given name and arity. */
  public void createPredicate(String predicate, int arity) {

    pn.insertPredicate(predicate, arity);
  }

  /** Creates a new predicate with given name and types for
      arguments. */
  public void createPredicate(String predicate, String argTypes[]) {

   createPredicate(predicate, argTypes.length);
  }


  /** Asserts a new axiom to the KB. */
  public void storeAxiom(Axiom anAxiom) {

  }


  /* *********** Interface specific functions ****************** */

  public void displayMessage(String msg) {

    // System.out.println(msg);
  }

  /* *********************************************************** */

  /** Returns the number of answers that exist for a given Query,
    without actually retrieving the answers from the server. */
  private int countQueryAnswers(Query query) {

    String[]  varList;
    int       numVars;
    String    parkaQuery;
    int       numAnswers = 0;
    
    varList = query.getVariables();
    numVars = varList.length;
    if (numVars == 0) {
      System.out.println("ERROR: Query has no variables");
      numAnswers = 0;
    }
    else {
      parkaQuery = formatSentence(query);
      pn.submitQuery(parkaQuery);
      numAnswers = pn.getNumTotalAnswersToQuery();
      if (numAnswers < 0)
	numAnswers = 0;
    }
    return numAnswers;
  }

  /** Given a Parka query string with a single predicate and variable,
    submits the query to the KB and, if sucessful, returns an array of 
    answers. It is up to the caller to make sure the query only has one
    variable. */
  private String[] submitSimpleQuery (String predicate,
				      String arg1, String arg2) {

    String    query;
    int       numAnswers;
    String[]  answers;

    query = formatAssert(predicate, arg1, arg2);
    // System.out.println("Query: " + query);
    pn.submitQuery(query);
    numAnswers = pn.getNumTotalAnswersToQuery();
    if (numAnswers <= 0) {
      // System.out.println("No answers returned");
      answers = new String[0];
    }
    else {
      /* pass start and end of requested set of answers (each set of
	 variable bindings count as one answer */
      answers = pn.getAnswers(1, numAnswers);
    }
    return answers;
  }


  /** Submit a query with a requested order for the return of answers.
    The answers will be returned in a ResultSet object. The order of
    the columns is specified by the array varOrder which consists of
    variable names. The maximum number of answers to return is
    specified in the third argument. */
  private ResultSet submitQuery (String query, String[] varOrder,
				 int numAnswersRequested) {

    int       numAnswers;
    String[]  vars;
    String[]  answers;
    ResultSet rset;

    // System.out.println("Query: " + query);
    pn.submitQuery(query);
    numAnswers = pn.getNumTotalAnswersToQuery();
    if (numAnswers <= 0) {
      // System.out.println("No answers returned");
      rset = new ResultSet(varOrder, 0);
    }
    else {
      // get variable names
      /* getVariables() is already called by pn.submitQuery(), we don't
       need to make another call to the server. */
      // vars = pn.getVariables();
      vars = pn.getVariableList();

      /* determine how many answers to request */
      if (numAnswersRequested != Query.ALL_ANSWERS)
	if (numAnswers > numAnswersRequested)
	  numAnswers = numAnswersRequested;

      /* pass start and end of requested set of answers (each set of
        variable bindings count as one answer)  */
      answers = pn.getAnswers(1, numAnswers);
      rset = getBindings(varOrder, vars, answers);
    }
    return rset;
  }


  /** Given an order of bindings, an array of variables in the query, and
     an array of the answers, places the answers in a 2-D array such that
     each row is a new set of bindings, consisting of the bindings in the
     order requested by the binds variable */
  ResultSet getBindings(String[] order, String[] vars, String[] answers) {

    int       numVars;
    int       numAnswers;
    int       numRows;
    int[]     bindpos;
    ResultSet rset;
    String[]  row;
    boolean   found = false;
    int       i, j;

    numVars = vars.length;
    numAnswers = answers.length;
    numRows = numAnswers / numVars;
    bindpos = new int[order.length];
    rset = new ResultSet(order, numRows);

    // System.out.println("Answers=" + numAnswers + ", Rows=" + numRows);
    /* determine the position of each bind variable in the answer */
    for (i=0; i < order.length; i++) {
      j = 0;
      found = false;
      while (j < numVars && found == false) {
	if (order[i].equals(vars[j]))
	  found = true;
	else
	  j++;
      }
      if (found == true)
	bindpos[i] = j;
      else
	bindpos[i] = -1;
    }
    for (i=0; i < numRows; i++) {
      row = new String[order.length];
      for (j=0; j < order.length; j++) {
	if (bindpos[j] >= 0)
	  row[j] = answers[i * numVars + bindpos[j]];
	else
	  row[j] = "";
      // System.out.println("rset[" + i + "][" + j + "] = " + rset[i][j]);
      }
      rset.addResult(row);
    }
    return rset;
  }


  /* Static functions */

  /** This function takes a string intended as an argument and escapes those
     characters that have special meaning in a Parka Assertion or query.
     In particular, the hash ('#'), backslash ('\') and parentheses ('(', ')')
     are escaped. Also, newlines are converted to spaces so that the 
     query string can be safely used in the file format */
  public static String escapeSpecialChars(String inString) {

    StringBuffer outString;
    int lastEsc = 0;
    int i;
    char c;

    outString = new StringBuffer(inString.length());
    for (i=0; i<inString.length(); i++) {
      c = inString.charAt(i);
      if (c == '#' || c =='\\' || c=='(' || c==')') {
	outString.append(inString.substring(lastEsc, i));
	outString.append("\\");         // insert a backslash before c
	lastEsc = i;
      }
      if (c == '\n') {
	outString.append(" ");
	lastEsc = i + 1;
      }
    }
    outString.append(inString.substring(lastEsc));
    return outString.toString();
  }

  /** Create a Parka assertion string. Pounds (#'s) are inserted after
    the predicate name and between each argument. Any special characters
    are escaped. If an argument is missing, it is replaced with a dash. */
  public static String formatAssert(String predicate, String args[]) {

    String assert;
    int i;

    assert = "(" + predicate;
    for (i=0; i < args.length; i++) {
      if (args[i] == null)
	args[i] = "-";
      assert = assert + "#" + escapeSpecialChars(args[i]);
    }
    assert = assert + ")";
    return assert;
  }

  public static String formatAssert(String predicate, String arg1, String arg2)
  {
    String assert;
    String args[];

    args = new String[2];
    args[0] = arg1;
    args[1] = arg2;
    assert = formatAssert(predicate, args);
    return assert;
  }

  /** Given a generic atom object, returns the equivalent Parka query
    or assertion string. */
  public static String formatAtom(Atom atom) {
    
    String   parkaAtom;
    String   predicate;
    String[] args;
    int i;

    predicate = atom.getPredicate();
    args = atom.getArguments();

    parkaAtom = "(" + predicate;
    for (i=0; i < args.length; i++) {
      if (args[i] == null)
	args[i] = "-";
      parkaAtom = parkaAtom + "#" + escapeSpecialChars(args[i]);
    }
    parkaAtom = parkaAtom + ")";
    return parkaAtom;
  }  

  /** Given a generic Sentence object, returns the equivalent Parka 
    string. */
  public static String formatSentence(Sentence sentence) {

    StringBuffer parkaString;
    Enumeration enum;
    Atom atom;
    String parkaAtom;

    parkaString = new StringBuffer();
    enum = sentence.getAtoms().elements();
    while (enum.hasMoreElements()) {
      atom = (Atom)enum.nextElement();
      parkaAtom = formatAtom(atom);
      parkaString.append(parkaAtom);
    }
    return parkaString.toString();
  }
}
