/* Copyright (C) 1999 Jeff Heflin - University of Maryland
 *	
 * This software is provided under the terms of the GNU Lesser General Public
 * License Version 2.1, February 1999. See the header in ShoeKb.java
 * and the file License.html for details.
 */

/* Change History:
   */

package ShoeKb;

import java.util.*;
import java.io.PrintStream;
import SHOE.*;
import ShoeXsb.XsbContextKb;

/** A subclass of ShoeKb that implements particular operations for
  KBs based on sets of axioms. Handles the locality of inference
  rules, backward-compatibility, and the semantics of typed relation
  arguments (i.e., an argument type implies that all instances in 
  that argument position are of the correct type). For this to work,
  the KBInterface used must provide the appropriate infrastructure
  for contexts, including the context predicate, definition of a
  query(C,V,Q,L) predicate, and appropriate construction of a
  contextualized query. */
public class ShoeAxiomaticKb extends ShoeKb
{

  private final static String CONTEXT_PRED = "context";

  /** Constructs a new SHOE KB interface using the specified API and
   ontology manager. Any messages reported by this object will be
   written to the specified log. */
  public ShoeAxiomaticKb(KBInterface kb, OntManager ontman,
			 PrintStream logOut) {

    super(kb, ontman, logOut);
  }

  /** Constructs a new SHOE KB interface using the specified API and
   ontology manager. Any messages reported by this object will be
   written to the specified log. */
  public ShoeAxiomaticKb(KBInterface kb, OntManager ontman, PrintStream logOut,
			 boolean implicitClaims) {

    // we alway have implicitClaims false, b/c axioms will be used to do
    // this
    super(kb, ontman, logOut, false);
  }

  /** Creates a new KB with the name supplied by the user. The
   necessary meta structures for cataloging ontology information are
   also created. */
  public void createKB(String kbName)
       throws KBException
    {
      // it is expected that the KBInterface createKB() method will
      // set up the necessary stuff for handling contexts, including
      // creating the context predicate and defining a special query
      // predicate
      super.createKB(kbName);
    }

  /** Issues an arbitrary query to the KB using the default
    ontology. Variables arguments are preceded by question marks. The
    answers are returned in a matrix. */
  public ResultSet issueQuery(Query query) {

    Query kbQuery;
    ResultSet rset;
    String curContext;

    curContext = getContextName(defaultOntId, defaultOntVer);
    kbQuery = toKbQuery(query);
    rset = ((XsbContextKb)kb).queryKB(kbQuery, curContext);
    return rset;
  }



  /* ***************** Protected Methods ********************* */

  /** Update the KB to make the given ontology backwards compatible
    with any appropriate prior versions. In a KB with axioms, this is
    done by creating axioms that map each term from the earlier
    version to the equivalent term in the latest version. */
  protected void setBackwardCompatibility(Ontology ont)
  {
    String ontId;
    String curContext;
    String   curVerNum;
    String[] compatVers;
    String   priorVerNum;
    Ontology priorOnt;
    Enumeration enum;
    SHOE_Obj shoeObj;
    String    priorObjName = null;
    String    curObjName = null;
    String[]  args = null;
    int       arity;
    Axiom     axiom = null;

    ontId = ont.getId();
    curVerNum = ont.getVersion();
    curContext = getContextName(ontId, curVerNum);
    compatVers = ont.getVersionsCompatibleWith();
    for (int i=0; i < compatVers.length; i++) {
      priorVerNum = compatVers[i];

      // Need a better way to find the previous version of an ontology.
      // If it has not already be loaded in OntManager, then we don't
      // know where to look for it, unless there is a proxy file that
      // specifies a URL.
      priorOnt = ontman.getOntology(ontId, priorVerNum);
      if (priorOnt == null) {
	System.out.println("Could not find prior ontology verison: " +
			   ontId + ", v." + priorVerNum);
	return;
      }

      // if prior version of ontology is not already in the KB, store
      // it and add to catalog
      String key = Ontology.makeKey(ontId, priorVerNum);
      if (storedOnts.containsKey(key) == false) {
	/* storeOntology() will add a catalog entry for the ontology */
	storeOntology(priorOnt);
      }

      // create axioms to match each vocabulary term (cats and rels)
      enum = priorOnt.getElements();
      while (enum.hasMoreElements()) {
	shoeObj = (SHOE_Obj)enum.nextElement();
	priorObjName = getUniqueName(ontId, priorVerNum, shoeObj.getKey());
	curObjName = getUniqueName(ontId, curVerNum, shoeObj.getKey());
	  
	if (curObjName == null) {
	  System.out.println(ontId + ",v." + curVerNum + " not compatible " +
			     "with version " + priorVerNum + " element " +
			     shoeObj.getKey());
	}
	else if (shoeObj instanceof Def_Category ||
	    shoeObj instanceof Def_Relation) {

	  /* What about renames, consts, and type? If v.1 defined a rename R
	     from ont O for term T, then how do we import the fact that
	     data which references v.1 R is valid vocabulary in BCompat v.2?
	     The problem is we currently resolve the name to its home
	     ontology and don't maintain the user's term! Thus, we can't
	     simply say all of the T's are v.2 R's, because some data might
	     use T as is.

	     Consider name in cs v.1.0 which renames base:name. Fortunately,
	     this is not a problem b/c cs v.1.0 also renames base:name to
	     name. Other case we won't be so lucky.
	     */
	    
	  if (shoeObj instanceof Def_Category) {
	    /*
	    // if categories are unary predicates
	    args = new String[1];
	    args[0] = "?X";
	    axiom = new Axiom();
	    axiom.addAntecedent(new Atom(CONTEXT_PRED, curContext));
	    axiom.addAntecedent(new Atom(priorObjName, args));
	    axiom.addConsequent(new Atom(curObjName, args));
	    */

	    // if categories are first class objects
	    // e..g, instance(c,X) :- context(ont1), instance(p,X).
	    axiom = new Axiom();
	    axiom.addAntecedent(new Atom(CONTEXT_PRED, curContext));
	    axiom.addAntecedent(new Atom(KBInterface.P_INSTANCE,
					 priorObjName, "?X"));
	    axiom.addConsequent(new Atom(KBInterface.P_INSTANCE,
					 curObjName, "?X"));
	  }
	  else if (shoeObj instanceof Def_Relation) {
	    arity = ((Def_Relation)shoeObj).getArity();
	    args = new String[arity];
	    for (int j=0; j < arity; j++)
	      args[j] = "?X" + (j+1);
	    axiom = new Axiom();
	    axiom.addAntecedent(new Atom(CONTEXT_PRED, curContext));
	    axiom.addAntecedent(new Atom(priorObjName, args));
	    axiom.addConsequent(new Atom(curObjName, args));
	  }

	  /* NOTE: make sure that axiom is asserted dynamically 
	     (use Prolog assert) */
	  kb.storeAxiom(axiom);
	}
      }
    }
  }

  /** Update the KB with information about ontologies extended by
    the given ontology. In a context based system, this can be used
    to provide axioms that relate contexts. */
  protected void setExtendedOnts(Ontology ont)
  {
    String curContext;
    String extContext;
    Use_Ontology[] useOnts;
    Use_Ontology aUseOnt;
    Axiom axiom;

    curContext = getContextName(ont.getId(), ont.getVersion());
    useOnts = ont.getUseOnts();
    for (int i=0; i < useOnts.length; i++) {
      aUseOnt = useOnts[i];
      extContext =  getContextName(aUseOnt.getId(), aUseOnt.getVersion());
      axiom = new Axiom(new Atom(CONTEXT_PRED, curContext),
			new Atom(CONTEXT_PRED, extContext));

   /* NOTE: make sure that axiom is asserted dynamically (use Prolog assert) */
      ((XsbContextKb)kb).storeDynamicAxiom(axiom);
    }

  }

  /** Given a SHOE DEF_CATEGORY and its source ontology, stores it in
    the KB. The isa links and names are stored. For the isa links,
    rules are used so that they only fire in the appropriate contexts. */
  protected void storeDefCategory(Def_Category dcat, Ontology ont) {

    String objKbName;               // the KB specific name of a SHOE object
    Enumeration pElmtEnum;
    Unique_Elt elmt_id;
    String parKbName;             // the KB specific name of a parent category
    String key;
    String curContext;
    Axiom axiom;

    key = ont.getKey();

    // System.out.println("Storing category " + dcat.getName());
    objKbName = getUniqueName(ont, dcat.getName());

    kb.createClass(objKbName);

    // store short name of category
    // disable the following condition to run OKBC version (can't use name
    // relation until it is loaded from the base ontology)
    if (dcat.getPhrase() != null)     
      kb.storeClaim(key, kbNameRel, objKbName, dcat.getPhrase());

    // store structural links to parents
    curContext = getContextName(ont.getId(), ont.getVersion());
    pElmtEnum = dcat.getParentIds();
    while (pElmtEnum.hasMoreElements()) {
      elmt_id = (Unique_Elt)pElmtEnum.nextElement();
      parKbName = getUniqueName(elmt_id);
      // store structural link

      /*
      // if categories are unary predicates
      // e.g., p(X) :- context(ont1), c(X)
      axiom = new Axiom();
      axiom.addAntecedent(new Atom(CONTEXT_PRED, curContext));
      axiom.addAntecedent(new Atom(objKbName, "?X"));
      axiom.addConsequent(new Atom(parKbName, "?X"));
      */

      // if categories are first class objects
      // e.g., isa(p,c) :- context(ont1) 
      axiom =  new Axiom();
      axiom.addAntecedent(new Atom(CONTEXT_PRED, curContext));
      axiom.addConsequent(new Atom(KBInterface.P_ISA, objKbName, parKbName));

      kb.storeAxiom(axiom);
      // kb.storeClaim(key, KBInterface.P_ISA, objKbName, parKbName);
    }    
  }


  /** Given a SHOE DEF_RELATION and its source ontology, stores it in
    the KB. A new predicate of the correct arity is created, and
    rule for infering the type of an instance (based on the types
    of the arguments that it appears in) are added. */
  protected void storeDefRelation(Def_Relation drel, Ontology ont) {

    String objKbName;               // the KB specific name of a SHOE object
    String key;
    Def_Arg[] dargs;
    String[] argTypes;
    Unique_Elt argTypeId;

    key = ont.getKey();

    objKbName = getUniqueName(ont, drel.getName());
    
    // kb.createPredicate(objKbName, drel.getArity());
    
    dargs = drel.getDefArgs();
    argTypes = new String[dargs.length];
    for (int i=0; i<dargs.length; i++) {
      argTypeId = dargs[i].getTypeId();
      argTypes[i] = getUniqueName(argTypeId);
    }
    kb.createPredicate(objKbName, argTypes);

    /* infer types from relation arguments here! If the type is a category,
       then all instances in that argument position must be a member of the
       category */
    String     curContext;
    int        arity;
    String[]   argVars;
    Unique_Elt typeId;
    String     typeKbName;
    Axiom      axiom;
    
    curContext = getContextName(ont.getId(), ont.getVersion());
    arity = drel.getArity();
    argVars = new String[arity];
    for (int i=0; i < arity; i++)
      argVars[i] = "?X" + (i+1);
    for (int i=0; i < dargs.length; i++) {
      if (dargs[i].getTypeClass() == Def_Arg.CATEGORY) {
	typeId = dargs[i].getTypeId();
	typeKbName = getUniqueName(typeId);
	
	/*
	// if categories are unary predicates
	axiom = new Axiom();
	axiom.addAntecedent(new Atom(CONTEXT_PRED, curContext));
	axiom.addAntecedent(new Atom(objKbName, argVars));
	axiom.addConsequent(new Atom(typeKbName, "?X" + (i+1)));
	*/

	// if categories are first class objects
	axiom = new Axiom();
	axiom.addAntecedent(new Atom(CONTEXT_PRED, curContext));
	axiom.addAntecedent(new Atom(objKbName, argVars));
	axiom.addConsequent(new Atom(KBInterface.P_INSTANCE, 
				     typeKbName, "?X" + (i+1)));

	kb.storeAxiom(axiom);
      }
    }
  }


  /** Given a SHOE DEF-INFERENCE rule and its source ontology, asserts it
    to the KB. */
  protected void storeInferenceRule(Def_Inference dinf, Ontology ont) {

    Axiom rule;
    Enumeration ants;
    Enumeration cons;
    SHOE_Obj obj;
    Atom conjunct;
    String ontContext;

    // System.out.println("Storing DEF-INFERENCE: " + dinf.toString());
    rule = new Axiom();

    // add contextual argument
    ontContext = getContextName(ont.getId(), ont.getVersion());
    rule.addAntecedent(new Atom(CONTEXT_PRED, ontContext));

    ants = dinf.getInfIf().elements();
    while (ants.hasMoreElements()) {
      obj = (SHOE_Obj)ants.nextElement();
      conjunct = shoeObjToAtom(obj);
      if (conjunct != null)
	rule.addAntecedent(conjunct);     
      else 
	System.out.println(obj.toString() + " resulted in a null conjunct");
    }
    cons = dinf.getInfThen().elements();
    while (cons.hasMoreElements()) {
      obj = (SHOE_Obj)cons.nextElement();
      conjunct = shoeObjToAtom(obj);
      if (conjunct != null) {
	rule.addConsequent(conjunct);     

	/*
	// we need to perform static analysis to determine if this axiom
	// might result in any programs that wouldn't terminate under
	// normal Prolog. However, we'll take the easy way out and
	// just table each consequent of each axiom that is a relation.
	String consPred = conjunct.getPredicate();
	if (!consPred.equals(P_ISA) && !consPred.equals(P_INSTANCE)) {
	  String[] consArgs = conjunct.getArguments();
	  ((XsbContextKb)kb).tablePredicate(consPred, consArgs.length);
	}
	*/
      }
      else 
	System.out.println(obj.toString() + " resulted in a null conjunct");
    }
    kb.storeAxiom(rule);
  }

  /* ********************** Private Methods ********************* */

  private String getContextName(String ontId, String ontVersion)
  {
    int seqNum;
    String contextName;

    seqNum = getOntSeqNumber(ontId, ontVersion);
    contextName = "ont" + seqNum;
    return contextName;
  }

  /** Given a list of SHOE variable names, turn them into Prolog variables
    and returns them as a Prolog list. */
  private String toPrologVariableSpec(String[] vars) {
    
    StringBuffer    varString;
    String          prologVar;

    varString = new StringBuffer("(");
    for (int i=0; i < vars.length; i++) {
      prologVar = vars[i].substring(1).toUpperCase();
      varString.append(prologVar);
      if (i < vars.length - 1)
	varString.append(",");
    }
    varString.append(")");
    return varString.toString();
  }

}



